#!/usr/bin/env python3
""" validation and config generation - pretty standard shit """
import sys
import re
import configparser
import os
import requests


class Configuration:
    """configuration file class"""

    def __init__(self, config_file):
        self.conf = configparser.ConfigParser()
        self.config_file = config_file
        if not os.path.exists(config_file):
            print(
                f"Error: Configuration file '{config_file}' not found. Creating blank config."
            )
            try:
                self.conf["Trakt"] = {
                    "client_id": "",
                    "client_secret": "",
                    "username": "",
                    "redirect_uri": "",
                    "oauth2_token": "",
                    "oauth2_refresh": "",
                }
                self.conf["Radarr"] = {
                    "url": "",
                    "api_key": "",
                    "trakt_list": "",
                    "trakt_list_privacy": "",
                }
                self.conf["Sonarr"] = {
                    "url": "",
                    "api_key": "",
                    "trakt_list": "",
                    "trakt_list_privacy": "",
                }
                with open(config_file, "w", encoding="utf-8") as configfile:
                    self.conf.write(configfile)
                    print(
                        "Please configure for oauth, use the -o=CODE parameter "
                        "and valid config credentials."
                    )
            except configparser.Error as error:
                print("An error occurred:", error)
            sys.exit(1)
        else:
            try:
                self.conf.read(config_file)
            except configparser.Error as error:
                print(f"Error occurred while reading the configuration file: {error}")
                sys.exit(1)

    def get_oauth(self, args, refresh=False):
        """gets the oauth token via refresh or code"""
        try:
            client_id = self.conf.get("Trakt", "client_id")
            client_secret = self.conf.get("Trakt", "client_secret")
            redirect_uri = self.conf.get("Trakt", "redirect_uri")
            if args.oauth:
                authorization_code = args.oauth
            if refresh or args.refresh:
                authorization_code = self.conf.get("Trakt", "oauth2_refresh")
                if authorization_code is None:
                    print(
                        "Trakt.tv Error: Exchanging refresh token failed. "
                        "You do not have a valid refresh_token in your config. "
                        "Please use -o instead."
                    )
                    sys.exit(1)
            if (
                (len(authorization_code) != 64)
                or (len(client_id) != 64)
                or (len(client_secret) != 64)
            ):
                print(
                    "You need to set and provide a valid code, client_id, and client_secret. "
                    "Please double check, and rerun the oauth command."
                )
                sys.exit(1)
            if not re.match(
                r"^(?:https?://)?(?:[-\w.]+)+(?::\d+)?(?:/.*)?$", redirect_uri
            ):
                print(
                    "You need to set the redirect_uri value to match trakt "
                    "and rerun the oauth command."
                )
                sys.exit(1)
        except configparser.Error as error:
            print(f"Error occurred while reading the configuration values: {error}")
            sys.exit(1)

        oauth_request = {
            "code": authorization_code,
            "client_id": client_id,
            "client_secret": client_secret,
            "redirect_uri": redirect_uri,
            "grant_type": "authorization_code",
        }
        if args.refresh or refresh:
            oauth_request["grant_type"] = "refresh_token"
            oauth_request["refresh_token"] = authorization_code
        try:
            response = requests.post(
                "https://api.trakt.tv/oauth/token",
                json=oauth_request,
                headers={"Content-Type": "application/json"},
                timeout=10,
            )
            response.raise_for_status()
            print("Authorization Code: ", authorization_code)
            print("Access Token: ", response.json().get("access_token"))
            print("Refresh Token: ", response.json().get("refresh_token"))
            self.conf.set("Trakt", "oauth2_token", response.json().get("access_token"))
            self.conf.set(
                "Trakt", "oauth2_refresh", response.json().get("refresh_token")
            )
            with open(self.config_file, "w", encoding="utf-8") as configfile:
                self.conf.write(configfile)
                print(
                    "Your configuration file was successfully updated "
                    "with your access/refresh token.\n"
                )
            if args.refresh is False and refresh is False:
                sys.exit(1)
            else:
                return response.json().get("access_token")
        except requests.exceptions.RequestException as error:
            print(error)
            print("Check your configuration, make sure they match Trakt.tv exactly")
            print(
                "Further Information: https://trakt.docs.apiary.io/#introduction/status-codes"
            )
            sys.exit(1)

    def validate_trakt_credentials(self):
        """validates trakt.tv credentials"""
        try:
            oauth2_bearer = self.conf.get("Trakt", "oauth2_token")
            trakt_api_key = self.conf.get("Trakt", "client_id")
            trakt_secret = self.conf.get("Trakt", "client_secret")
            user = self.conf.get("Trakt", "username")
        except configparser.Error as error:
            print(f"Error occurred while reading the configuration values: {error}")
            sys.exit(1)
        if len(user) == 0:
            print(
                "Error: Invalid configuration values. "
                "[Trakt] username should not be empty."
            )
            sys.exit(1)
        # validate the lengths of all the keys are correct
        if (
            len(oauth2_bearer) != 64
            or len(trakt_api_key) != 64
            or len(trakt_secret) != 64
        ):
            print(
                "Error: Invalid configuration values. "
                "[Trakt] oauth2_token/client_id/client_secret "
                "should all have lengths of 64 characters."
            )
            print(
                "Run with -o parameter with accurate username, redirect_uri, "
                "client_id, and client_secret values set in config."
            )
            sys.exit(1)
        return oauth2_bearer, trakt_api_key, user, trakt_secret

    def validate_arr_configuration(self, arr_api, trakt_api, arr, args):
        """validates the specified arr config"""
        try:
            if not args.list:
                trakt_api.list = None
            if not args.privacy:
                trakt_api.list_privacy = None

            arr_api.api_url = self.conf.get(arr, "url").rstrip("/")
            arr_api.api_key = self.conf.get(arr, "api_key")
            trakt_api.list_privacy = (
                self.conf.get(arr, "trakt_list_privacy")
                if (trakt_api.list_privacy is None)
                else trakt_api.list_privacy
            )
            trakt_api.list = (
                self.conf.get(arr, "trakt_list")
                if (trakt_api.list is None)
                else trakt_api.list
            )
        except configparser.Error as error:
            print(f"Error occurred while reading the configuration values: {error}")
            sys.exit(1)
        if not re.match(
            r"^(?:https?://)?(?:.+:.+@)?(?:[-\w.]+)+(?::\d+)?(?:/.*)?$", arr_api.api_url
        ):
            print(
                f"Error: Invalid configuration value. [{arr}] 'url' does not match a URL pattern."
            )
            sys.exit(1)
        if len(arr_api.api_key) != 32:
            print(
                f"Error: Invalid configuration values. "
                f"[{arr}] api_key should have lengths of 32 characters."
            )
            sys.exit(1)



#!/usr/bin/env python3
""" handles the arr api calls and requests """
import sys
import requests
from urllib.parse import urlparse


class ArrAPI:
    """arr api handler class"""

    def __init__(self):
        self.api_url: str
        self.api_key: str
        self.endpoint = {
            "Sonarr": ("series", "tvdb", "shows"),
            "Radarr": ("movie", "tmdb", "movies"),
        }

    # queries arr and gets the return from the end point passed to it
    def arr_get(self, arr, endpoint, timeout):
        """sends the get request to the arr endpoint"""
        try:
            parsed_url = urlparse(self.api_url)

            url_host = parsed_url._replace(
                netloc=parsed_url.netloc.split("@")[-1]
            ).netloc

            url_path = parsed_url.path if parsed_url.path is not None else ""

            request_url = f"{parsed_url.scheme}://{url_host}{url_path}"

            if (parsed_url.username is None) or (parsed_url.password is None):
                response = requests.get(
                    f"{request_url}/api/v3/{endpoint}",
                    params={"apikey": self.api_key},
                    timeout=timeout,
                )
            else:
                response = requests.get(
                    f"{request_url}/api/v3/{endpoint}",
                    params={"apikey": self.api_key},
                    timeout=timeout,
                    auth=(parsed_url.username, parsed_url.password),
                )
            response.raise_for_status()
            return response
        except requests.exceptions.ConnectTimeout:
            print(f"{arr}: Connection Timed Out. Check your URL.")
            sys.exit(1)
        except requests.exceptions.ConnectionError as error:
            print(f"{arr}: Connection Error. Check Your URL or server.")
            error = str(error).split("] ")[1].split("'")[0]
            print(f"{arr}: {error}")
            sys.exit(1)
        except requests.exceptions.HTTPError as error:
            if "401" in str(error):
                print(
                    f"{arr} Error: API key incorrect. "
                    "Please double check your key and config file."
                )
                sys.exit(1)
            print(f"{arr} Error:")
            print(f"{arr}: {error}")
            sys.exit(1)

    def get_id(self, arr, search_term, endpoint, term):
        """sends a request to get get necessary ids"""
        response = self.arr_get(arr, endpoint, 10)

        # creates a dict for the term: id
        id_dict = {item[term]: item["id"] for item in response.json()}

        # if it can't find an id for the term error and exit
        if id_dict.get(search_term) is None:
            print(f"{arr} Error: No matching {endpoint if (endpoint != "qualityprofile") else "quality profile"} found.")
            sys.exit(1)

        # return the id
        return id_dict.get(search_term)

    def get_list(self, args, arr):
        """sends the get request to the movies/series arr endpoint"""
        response = self.arr_get(arr, f"{self.endpoint[arr][0]}", 10)
        arr_data = {}

        # parsing out the tmdb/tvdb/imdb id's,
        #  monitored status, quality profile id, title, and any tags
        for item in response.json():
            arr_data[item[f"{self.endpoint[arr][1]}Id"]] = [
                item.get("imdbId"),
                item.get("monitored"),
                item.get("qualityProfileId"),
                item.get("title"),
                item.get("tags"),
                item.get("hasFile") if (arr == "Radarr") else None
            ]
        arr_ids = list(arr_data.keys())

        # if its monitored, add to arr ids
        if args.mon:
            arr_ids = [key for key, value in arr_data.items() if value[1]]

        # get the current filtered arr_ids that qualify for the specified quality profile
        if args.qualityprofile:
            qp_id = self.get_id(arr, args.qualityprofile, "qualityprofile", "name")
            arr_ids = list(
                filter(lambda item: arr_data.get(item, [None])[2] is qp_id, arr_ids)
            )

        # same as above, but for tags
        if args.tag:
            tag_id = self.get_id(arr, args.tag, "tag", "label")
            arr_ids = list(
                filter(lambda item: tag_id in arr_data.get(item, [None])[4], arr_ids)
            )
        if arr == "Radarr" and args.missing:
            arr_ids = list(
                filter(lambda item: arr_data.get(item, [None])[5] == False, arr_ids)
            )
        # if imdb id is in arr, add it to the imdb id list
        arr_imdb = [
            value[1][0]
            for value in arr_data.items()
            if value[1][0] is not None and value[0] in arr_ids
        ]

        return arr_ids, arr_imdb, arr_data



#!/usr/bin/env python3
""" main script, arguments and executions """
import sys
import argparse
from os import name, path, getenv

from retraktarr.api.arr import ArrAPI
from retraktarr.api.trakt import TraktAPI
from retraktarr.config import Configuration


def main():
    """main entry point defines args and processes stuff"""
    try:
        with open(
            path.join(path.dirname(path.abspath(__file__)), "VERSION"), encoding="utf-8"
        ) as f:
            VERSION = f.read()
    except OSError as e:
        VERSION = "MISSING"

    parser = argparse.ArgumentParser(
        description="Starr App -> Trakt.tv List Backup/Synchronization"
    )
    parser.add_argument(
        "--oauth",
        "-o",
        type=str,
        help="Update OAuth2 Bearer Token."
        " Accepts the auth code and requires valid Trakt "
        "config settings (ex: -o CODE_HERE)",
    )
    parser.add_argument(
        "--radarr",
        "-r",
        action="store_true",
        help="Synchronize Radarr movies with Trakt.tv",
    )
    parser.add_argument(
        "--sonarr",
        "-s",
        action="store_true",
        help="Synchronize Sonarr series with Trakt.tv",
    )
    parser.add_argument(
        "--all",
        "-all",
        "-a",
        action="store_true",
        help="Synchronize both Starr apps with Trakt.tv",
    )
    parser.add_argument(
        "--mon",
        "-m",
        action="store_true",
        help="Synchronize only monitored content with Trakt.tv",
    )
    parser.add_argument(
        "--missing",
        action="store_true",
        help="Synchronize only missing Radarr content with Trakt.tv",
    )
    parser.add_argument(
        "--qualityprofile",
        "-qp",
        type=str,
        help="The quality profile you wish to sync to Trakt.tv",
    )
    parser.add_argument(
        "--tag", "-t", type=str, help="The arr tag you wish to sync to Trakt.tv"
    )
    parser.add_argument(
        "--cat",
        "-c",
        action="store_true",
        help="Add to the Trakt.tv list without "
        "deletion (concatenate/append to list)",
    )
    parser.add_argument(
        "--list",
        "-l",
        type=str,
        help="Specifies the Trakt.tv list name. (overrides config file settings)",
    )
    parser.add_argument(
        "--wipe",
        "-w",
        action="store_true",
        help="Erases the associated list and performs a sync "
        "(requires -all or -r/s)",
    )
    parser.add_argument(
        "--privacy",
        "-p",
        type=str,
        help="Specifies the Trakt.tv list privacy settings "
        "(private/friends/public - overrides config file settings)",
    )
    parser.add_argument(
        "--refresh",
        action="store_true",
        help="Forces a refresh_token exchange (oauth) "
        "and sets the config to a new tokens.",
    )
    parser.add_argument(
        "--timeout",
        type=str,
        help="Specifies the timeout in seconds to use for " "POST commands to Trakt.tv",
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="Displays current version information",
    )
    parser.add_argument(
        "--config",
        action="store",
        nargs="?",
        const=True,
        default=None,
        help="If a path is provided, retraktarr will use this config file, otherwise it outputs default config location.",
    )
    args = parser.parse_args()
    print(f"\nretraktarr v{VERSION}")
    if args.version:
        sys.exit(0)

    if args.config is not True and args.config is not None:
        config_path = args.config
    else:
        config_path = (
            f'{path.expanduser("~")}{path.sep}.config{path.sep}retraktarr.conf'
        )
        if args.config is True:
            print(f"Current default config file is {config_path}")
            exit(0)

    print(f"Validating Configuration File: {config_path}\n")
    config = Configuration(config_path)
    if args.oauth:
        config.get_oauth(args)
    if args.refresh:
        config.get_oauth(args)

    (
        oauth2_bearer,
        trakt_api_key,
        trakt_user,
        trakt_secret,
    ) = config.validate_trakt_credentials()
    trakt_api = TraktAPI(oauth2_bearer, trakt_api_key, trakt_user, trakt_secret)
    if args.list:
        trakt_api.list = args.list
    if args.privacy:
        trakt_api.list_privacy = args.privacy
    if args.timeout:
        trakt_api.post_timeout = args.timeout
    if args.radarr or args.all or args.sonarr:
        arr_api = ArrAPI()

    if args.radarr or args.all:
        config.validate_arr_configuration(arr_api, trakt_api, "Radarr", args)
        tvdb_ids, tmdb_ids, imdb_ids, trakt_ids = trakt_api.get_list(
            args, arr_api.endpoint["Radarr"][0]
        )
        arr_ids, arr_imdb, arr_data = arr_api.get_list(args, "Radarr")
        print("[Radarr]")
        trakt_api.add_to_list(
            args,
            arr_api.endpoint["Radarr"][2],
            arr_data,
            tmdb_ids,
            arr_api.endpoint["Radarr"][1],
            imdb_ids,
            arr_ids,
            arr_imdb,
            trakt_ids,
        )
        print(f"Total Movies: {len(arr_ids)}\n")

    if args.sonarr or args.all:
        config.validate_arr_configuration(arr_api, trakt_api, "Sonarr", args)
        tvdb_ids, tmdb_ids, imdb_ids, trakt_ids = trakt_api.get_list(
            args, arr_api.endpoint["Sonarr"][2].rstrip("s")
        )

        arr_ids, arr_imdb, arr_data = arr_api.get_list(args, "Sonarr")
        print("[Sonarr]")
        trakt_api.add_to_list(
            args,
            arr_api.endpoint["Sonarr"][2],
            arr_data,
            tvdb_ids,
            arr_api.endpoint["Sonarr"][1],
            imdb_ids,
            arr_ids,
            arr_imdb,
            trakt_ids,
        )
        print(f"Total Series: {len(arr_ids)}")
        sys.exit(1)

    if args.radarr or args.all or args.refresh:
        sys.exit(1)

    parser.print_help()


if __name__ == "__main__":
    main()



import os

from setuptools import setup, find_packages

"""
setup file for package publishing
"""
# User-friendly description from README.md
current_directory = os.path.dirname(os.path.abspath(__file__))
try:
    with open(os.path.join(current_directory, "README.md"), encoding="utf-8") as f:
        LONG_DESCRIPTION = f.read()
except OSError as e:
    LONG_DESCRIPTION = ""

try:
    with open(
        os.path.join(current_directory, f"retraktarr{os.path.sep}VERSION"),
        encoding="utf-8",
    ) as f:
        VERSION_NO = f.read()
except OSError as e:
    VERSION_NO = ""


with open("requirements.txt") as reqs_file:
    requirements = reqs_file.read().splitlines()

setup(
    # Name of the package
    name="retraktarr",
    # Start with a small number and increase it with
    # every change you make https://semver.org
    # Chose a license from here: https: //
    # help.github.com / articles / licensing - a -
    # repository. For example: MIT
    license="MIT",
    version=VERSION_NO,
    # Short description of your library
    description=("a simple Arr -> Trakt.tv list sync script"),
    entry_points={"console_scripts": ["retraktarr = retraktarr:main"]},
    # Long description of your library
    install_requires=requirements,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    # long_description=long_description,
    # long_description_content_type="text/markdown",
    # Your name
    author="zakkarry",
    # Your email
    author_email="zak@ary.dev",
    # Either the link to your github or to your website
    url="https://github.com/zakkarry",
    # Link from which the project can be downloaded
    download_url="https://github.com/zakkarry/retraktarr",
    packages=find_packages(exclude=[".github"]),
    package_data={"retraktarr": ["VERSION"]},
)



#!/usr/bin/env python3
""" handles the trakt api lists and requests (add/delete/etc) """
import sys
import time
import json
import re
import requests
from retraktarr.config import Configuration


class TraktAPI:
    """trakt API handler class"""

    def __init__(self, oauth2_bearer, trakt_api_key, trakt_user, trakt_secret):
        self.oauth2_bearer = oauth2_bearer
        self.trakt_api_key = trakt_api_key
        self.user = trakt_user
        self.trakt_secret = trakt_secret
        self.list = ""
        self.json: dict
        self.response = None
        self.list_len = []
        self.list_privacy = "public"
        self.list_limit = 1000
        self.post_timeout = 60
        self.trakt_session = requests.Session()
        self.trakt_hdr = {
            "Content-Type": "application/json",
            "trakt-api-version": "2",
            "trakt-api-key": self.trakt_api_key,
            "Authorization": f"Bearer {self.oauth2_bearer}",
        }

    def refresh_header(self, oauth_token):
        """refresh header (used for oauth automatic refresh)"""
        self.trakt_hdr = {
            "Content-Type": "application/json",
            "trakt-api-version": "2",
            "trakt-api-key": self.trakt_api_key,
            "Authorization": f"Bearer {oauth_token}",
        }

    def normalize_trakt(self, input):
        # remove all parenthesis and brackets
        normalized = re.sub(r"[()\[\]]", "", input)
        # non-'-' and '_' characters with a hyphen
        normalized = re.sub(r"[^a-zA-Z0-9-_]", "-", normalized)
        # duplicate sequential hyphens
        normalized = re.sub(r"-+", "-", normalized)
        return normalized.strip("-")

    def get_trakt(self, path, args, media_type, timeout):
        """gets json response from the specified path for applicable media_type (show/movie)"""

        time.sleep(1)
        try:
            response = self.trakt_session.get(
                f"https://api.trakt.tv/{path}", headers=self.trakt_hdr, timeout=timeout
            )
            response.raise_for_status()
            if response.status_code != 200:
                print(
                    f"Trakt.tv Error: Unexpected status code return: {response.status_code}."
                )
                sys.exit(1)
            else:
                return response
        except requests.exceptions.ConnectTimeout:
            print("Trakt.tv Error: Connection Timed Out. Check your internet.")
            sys.exit(1)
        except requests.exceptions.ConnectionError as error:
            print("Trakt.tv: Connection Error. Check your internet.")
            print(f"{error}")
            sys.exit(1)
        except requests.exceptions.HTTPError as error:
            # checks if the list is missing mostly
            if "404" in str(error):
                return 404
            # checks if an oauth_refresh token is available
            # and if so assume that the token has expired and attempt a refresh automatically
            if "401" in str(error) or "400" in str(error) or "403" in str(error):
                config = Configuration("retraktarr.conf")

                # checks the config for the refresh, if exists, update the
                # header and rerun the command and return original intended results
                if (
                    config.conf.get("Trakt", "oauth2_refresh")
                    and self.oauth2_bearer
                    == self.trakt_hdr.get("Authorization").split(" ")[1]
                ):
                    print(
                        "Error: You may have a expired token. Attempting a refresh command."
                    )

                    # refresh the header with the new auth, update the config file with new tokens
                    self.refresh_header(config.get_oauth(args, True))
                    time.sleep(1)

                    # return the intended original results
                    return self.get_trakt(path, args, media_type, timeout=timeout)

                # no oauth_refresh token is available, error out.
                print(
                    "Error: You likely have a bad ClientID/Secret or expired/invalid token."
                    "\nPlease check your config and attempt the refresh or oauth command (-o) again"
                )
                sys.exit(1)
            print(
                f"Trakt.tv Error: Unexpected status code return: {response.status_code}."
            )
            sys.exit(1)

    def get_list(self, args, media_type):
        """ " gets the specified trakt list and settings (account limits)"""

        # grabs the users settings and sets the list limits
        response = self.get_trakt("users/settings", args, media_type, timeout=10)
        self.list_limit = (
            response.json().get("limits", {}).get("list", {}).get("item_count", None)
        )

        # sends a get request for the list and all of its items
        response = self.get_trakt(
            f"users/{self.normalize_trakt(self.user)}/lists/{self.normalize_trakt(self.list)}/items",
            args,
            media_type,
            timeout=30,
        )

        # returns empty lists if the list does not exist
        if response == 404:
            return [], [], [], []

        # sets all of the ids into lists for parsing/adding/logic
        self.list_len = [
            item.get("id", {}) for item in response.json() if item.get("id", {})
        ]

        tvdb_ids = [
            item.get(media_type, {}).get("ids", {}).get("tvdb")
            for item in response.json()
            if item.get(media_type, {}).get("ids", {}).get("tvdb") is not None
        ]

        tmdb_ids = [
            item.get(media_type, {}).get("ids", {}).get("tmdb")
            for item in response.json()
            if item.get(media_type, {}).get("ids", {}).get("tmdb") is not None
        ]

        imdb_ids = [
            item.get(media_type, {}).get("ids", {}).get("imdb")
            for item in response.json()
            if item.get(media_type, {}).get("ids", {}).get("imdb") is not None
        ]

        # makes a list of all trakt ids so we have every single item
        # guarenteed (we use this id for wiping)
        trakt_ids = [
            item.get("movie", {}).get("ids", {}).get("trakt")
            for item in response.json()
            if item.get("movie", {}).get("ids", {}).get("trakt") is not None
        ]
        for item in response.json():
            if item.get("show", {}).get("ids", {}).get("trakt") is not None:
                trakt_ids.append(item.get("show", {}).get("ids", {}).get("trakt"))

        # spit out the lists and json to main
        self.json = response.json()
        return tvdb_ids, tmdb_ids, imdb_ids, trakt_ids

    def post_trakt(self, path, post_json, args, media_type, timeout):
        """
        sends a post command to trakt
        post_json is json.dumps'd json, path is the url to append to the user url
        """

        time.sleep(1)
        try:
            response = self.trakt_session.post(
                f"https://api.trakt.tv/users/{self.normalize_trakt(self.user)}/{path}",
                headers=self.trakt_hdr,
                data=post_json,
                timeout=timeout if not args.timeout else self.post_timeout,
            )
            response.raise_for_status()
            if response.status_code in (200, 201, 204):
                return response
        except requests.exceptions.ConnectTimeout:
            print("Trakt.tv Error: Connection Timed Out. Check your internet.")
            sys.exit(1)
        except requests.exceptions.ReadTimeout:
            print(
                "Trakt.tv Error: Connection Timed Out Mid-Stream. Increase your --timeout. "
            )
            sys.exit(1)
        except requests.exceptions.ConnectionError as error:
            print("Trakt.tv: Connection Error. Check your internet.")
            print(f"{error}")
            sys.exit(1)
        except requests.exceptions.HTTPError as error:
            # http error parsing
            if "401" in str(error) or "403" in str(error):
                print(
                    "Trakt.tv Error: You likely have a bad OAuth2 Token, "
                    "username, or ClientID/API key.\n"
                    "Please check your config, revalidate with the oauth2 "
                    "command, and try again."
                )
                sys.exit(1)
            elif "420" in str(error):
                print(
                    "Trakt.tv Error:"
                    f"Your additions to ({self.trakt_list}) exceeds your item limits."
                    "You will need Trakt VIP."
                )
                sys.exit(1)
            elif "404" in str(error):
                # if the list doesn't exist, we create it
                # then rerun the same post commands
                # return the response as if nothing happened :)
                print(
                    "Trakt.tv Error (404): "
                    f"https://trakt.tv/users/{self.normalize_trakt(self.user)}/lists/{self.normalize_trakt(self.list)} not found...\n"
                )
                trakt_add_list = {
                    "name": self.list,
                    "description": "Created using retraktarr "
                    "(https://github.com/zakkarry/retraktarr)",
                    "privacy": self.list_privacy,
                    "allow_comments": False,
                }
                # adds the list
                self.post_trakt(
                    "lists", json.dumps(trakt_add_list), args, media_type, timeout=15
                )
                print(f"Creating {self.list_privacy} Trakt.tv list: ({self.list})...\n")
                time.sleep(1)

                # retry the POST and returns the intended original results
                response = self.post_trakt(
                    path,
                    post_json,
                    args,
                    media_type,
                    timeout=timeout if not args.timeout else self.post_timeout,
                )
                return response

    def del_from_list(
        self,
        args,
        media_type,
        arr_data,
        trakt_ids,
        idtag,
        trakt_imdb_ids,
        arr_ids,
        arr_imdb,
        all_trakt_ids,
    ):
        """
        super complicated logic to find and identify unneeded ids
        and remove them from the trakt list before adding
        """
        trakt_del, extra_imdb_ids, extra_ids, filtered_extra_imdb_ids, wrong_ids = (
            [] for i in range(5)
        )
        if not args.cat:  # not catenating to the list
            if len(all_trakt_ids) > 0:  # if the trakt list already has contents
                # if we're not wiping the entire list (--wipe)
                if not args.wipe:
                    needed_ids = set(arr_ids) - set(trakt_ids)
                    # set needed ids to all wanted ids from arr minus whats already on the list

                    extra_ids = set(trakt_ids) - set(arr_ids)
                    # all items from the list minus what is in arrs - tmdb/tvdb to be removed

                    extra_imdb_ids = set(trakt_imdb_ids) - set(arr_imdb)
                    # same as above, but specifically imdbids in case tmdb/tvdb is missing

                    # run through trakt's API json response for the list
                    # check the type of media and compare to what arr's media is
                    # see if the idtag (tvdb/tmdb) is missing
                    # and add the imdb id from trakt to a filtered extra's list
                    filtered_extra_imdb_ids = [
                        data[media_type.rstrip("s")]["ids"].get("imdb")
                        for data in filter(
                            lambda data: data["type"] == media_type.rstrip("s")
                            and data[media_type.rstrip("s")]["ids"].get(idtag) is None
                            and data[media_type.rstrip("s")]["ids"].get("imdb")
                            in extra_imdb_ids,
                            self.json,
                        )
                    ]

                    # check if there are extra tmdb/tvdb id's to be removed from trakt
                    # skip if wipe since we'd remove all
                    if len(extra_ids) > 0 and not args.wipe:
                        trakt_del = {media_type: []}
                        # create a dictionary for 'imdb' values in arr_data that could be wrong
                        arr_imdb_lookup = {
                            value[1][0]: value[0]
                            for value in arr_data.items()
                            if value[0] in needed_ids
                        }
                        # run through extra_ids
                        # check if the extra_id from trakt doesnt exist in arrs
                        # (almost always deleted/outdate trakt info)
                        for item in extra_ids:
                            if item not in arr_ids:
                                # if it isn't in our arr, append to the delete json
                                # we will readd if its on the needed_ids list later.
                                trakt_del[media_type].append({"ids": {idtag: item}})

                                # checks if the tmdb/tvdb is not in the arr's db
                                # ends up determining if the id is wrong and it will be readded
                                if item not in arr_data.keys():
                                    # run through the json
                                    for data in self.json:
                                        # check if the tvdb/tmdb id is in the arr
                                        if (
                                            data[media_type.rstrip("s")]["ids"].get(
                                                idtag
                                            )
                                            == item
                                        ) and (
                                            data[media_type.rstrip("s")]["ids"].get(
                                                "imdb"
                                            )
                                            in arr_imdb_lookup
                                        ):
                                            # if the imdb id is correct but tmdb/tvdb is not, add to wrong_ids
                                            wrong_ids.append(item)
                                            break
                    # remove the needed wrong ids from extra_ids so they dont delete
                    extra_ids = set(extra_ids) - set(wrong_ids)

                    # takes the filtered extra imdb ids
                    # (missing a corresponding tmdb/tvdb from arr on trakt)
                    if len(filtered_extra_imdb_ids) > 0 and not args.wipe:
                        # if there's no extra_ids (tvdb/tmdb)
                        # just build the entire json for imdb since it would be empty anyway
                        if len(extra_ids) == 0:
                            trakt_del = {
                                media_type: [
                                    {"ids": {"imdb": item}}
                                    for item in filtered_extra_imdb_ids
                                ]
                            }
                        # else, run through the filtered id's
                        # append the imdbid's without corresponding tvdb/tmdb in arr for removal
                        else:
                            for item in filtered_extra_imdb_ids:
                                trakt_del[media_type].append({"ids": {"imdb": item}})

                # if filtered is less is 0 or wipe, set needed_ids to all
                # build a json to delete everything (in case of wipe)
                # we wont run delete if the list is empty anyway...
                else:
                    needed_ids = set(arr_ids)
                    trakt_del = {
                        "shows": [{"ids": {"trakt": item}} for item in all_trakt_ids],
                        "movies": [{"ids": {"trakt": item}} for item in all_trakt_ids],
                    }
            else:
                needed_ids = set(arr_ids)
        else:
            needed_ids = set(arr_ids)
        # does some calculations on what the end list count would be
        # compares to your trakt list limits
        if (
            (
                len(self.list_len)
                + len(needed_ids)
                - len(extra_ids)
                - len(filtered_extra_imdb_ids)
            )
            > self.list_limit
        ) or (args.wipe and (len(needed_ids) > self.list_limit)):
            print(
                f"Error: Your additions to ({self.list}) exceeds your item limits."
                "You will need Trakt VIP."
            )
            sys.exit(1)

        # checks if there are extra ids to be removed
        # or list has items and a wipe was requested
        # since we removed wrong id's this wont be ran if there is nothing but wrong ids....
        if (len(extra_ids) > 0 or len(filtered_extra_imdb_ids) > 0) or (
            args.wipe and (len(all_trakt_ids) > 0)
        ):
            # sends the remove from list request
            self.post_trakt(
                f"lists/{self.normalize_trakt(self.list)}/items/remove",
                json.dumps(trakt_del),
                args,
                media_type,
                timeout=60,
            )

            # if not a wipe, display what was deleted...dont flood if wiping :P
            if not args.wipe:
                print(
                    f"Number of Deleted {media_type.title()}:  "
                    f"{len(extra_ids) + len(filtered_extra_imdb_ids)}"
                )

                # iterate through extra ids (tvdb/tmdb)
                # display the idname, the title, and the id itself
                for extra_id in extra_ids:
                    # uses the arr's json structure to display info
                    if extra_id in arr_data:
                        print(
                            f"        {idtag.upper()}: "
                            f"{arr_data.get(extra_id, [None])[3]} - {extra_id}"
                        )
                        continue

                    # if it can't grab the data from arr, it was deleted from the arr
                    # it will search trakt's json (slower) for the titles
                    for data in self.json:
                        if data[media_type.rstrip("s")]["ids"].get(idtag) == extra_id:
                            print(
                                f"        {idtag.upper()}: "
                                f"{data[media_type.rstrip('s')].get('title')} - {extra_id}"
                            )
                            break

                # same as above, but for imdb, but missing tvdb/tmdb on trakt
                # loops through filtered extra imdb's copy, removing as it finds matches,
                # so it doesnt double display since the data might be in both arr and trakt
                arr_data_items = arr_data.items()
                for item in filtered_extra_imdb_ids.copy():
                    for value in arr_data_items:
                        if item == value[1][0]:
                            print(f"        IMDB: {value[1][3]} - {value[1][0]}")
                            filtered_extra_imdb_ids.remove(item)
                            break
                    if item in filtered_extra_imdb_ids:
                        for data in self.json:
                            imdbid = data[media_type.rstrip("s")]["ids"].get("imdb")
                            if imdbid == item:
                                print(
                                    f"        IMDB: {data[media_type.rstrip('s')].get('title')}"
                                    f" - {imdbid}"
                                )
                                break
        return needed_ids

    def add_to_list(
        self,
        args,
        media_type,
        arr_data,
        trakt_ids,
        idtag,
        trakt_imdb_ids,
        arr_ids,
        arr_imdb,
        all_trakt_ids,
    ):
        """
        parse out and compares arr lists with trakt, then runs
        del_from_list and updates/adds to the list
        """

        trakt_add = {media_type: []}
        needed_ids = self.del_from_list(
            args,
            media_type,
            arr_data,
            trakt_ids,
            idtag,
            trakt_imdb_ids,
            arr_ids,
            arr_imdb,
            all_trakt_ids,
        )

        # build the add to list json, if imdb is not available just use tmdb/tvdb
        trakt_add = {
            media_type: [
                {"ids": {idtag: item, "imdb": arr_data.get(item, [None])[0]}}
                if arr_data.get(item, [None])[0] is not None
                else {"ids": {idtag: item}}
                for item in needed_ids
            ]
        }
        # sends the add to list request
        response = self.post_trakt(
            f"lists/{self.normalize_trakt(self.list)}/items",
            json.dumps(trakt_add),
            args,
            media_type,
            timeout=60,
        )

        # gets the count for the add results...
        added_items = response.json()["added"][media_type.lower()]
        listed_items = response.json()["list"]["item_count"]
        not_found_items = response.json()["not_found"].get(media_type.lower(), [])

        # sets the state for items that parsing what really wasnt added due to not corresponding ids
        real_not_found_items = []

        # goes through the ids of not found items on the add request
        for not_found_item in not_found_items:
            # grabs the tvdb/tmdb id of the not found item
            # we can assume it exists since its coming from arr
            idtag_value = not_found_item.get("ids", {}).get(idtag)

            # make sure its not a fucked up json response, and check that the response exists.
            # then append the real not found
            if idtag_value is not None and idtag_value in arr_data.keys():
                real_not_found_items.append(idtag_value)

        print(f"Number of {media_type.title()} Added: {added_items}")

        # if real not found is over 0, print the results from the arr_data
        if len(real_not_found_items) > 0:
            print(
                f"Number of {media_type.title()} Not Found: {len(real_not_found_items)}"
            )
            for real_not_found_item in real_not_found_items:
                print(
                    f"        {idtag.upper()}: "
                    f"{arr_data.get(real_not_found_item, [None])[3]} - {real_not_found_item}"
                )

        # if real not found is 0, finish up
        else:
            print(f"Number of {media_type.title()} Not Found: 0")
        print(f"Number of {media_type.title()} Listed: {listed_items}")

        # close the requests session
        self.trakt_session.close()
